Package: dynasim
Type: Package
Title: Dynamics Similarity Coefficient
Version: 1.1
Authors@R: c(
    person(given = "Chun-Xiao", family = "Nie", 
           email = "niechunxiao2009@163.com", 
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-7790-0803"))
  )
Description: Implements the quantile-graph based Dynamics Similarity Coefficient (DSC) for comparing intrinsic dynamics of time series.
License: GPL-3
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-04-22 14:41:49 UTC; niech
Author: Chun-Xiao Nie [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7790-0803>)
Maintainer: Chun-Xiao Nie <niechunxiao2009@163.com>
Repository: CRAN
Date/Publication: 2026-04-23 20:00:09 UTC
Built: R 4.5.2; ; 2026-04-24 13:21:33 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cross_transition_matrix)
export(discretize_quantile)
export(dsc)
export(dsc_matrix)
export(gdsc)
export(gdsc_matrix)
export(mdsc)
export(transition_matrix)

# DynaSim

<!--- These are examples. See https://shields.io for others or to customize this set of shields. You might want to include dependencies, project status and licence info here --->
![GitHub repo size](https://img.shields.io/github/repo-size/MarcusHA94/dynasim)
![GitHub contributors](https://img.shields.io/github/contributors/MarcusHA94/dynasim)

<!-- Project name is a `<utility/tool/feature>` that allows `<insert_target_audience>` to do `<action/task_it_does>`. -->

The dynasim package simulates dynamic systems in the form:
```math
\mathbf{M}\ddot{\mathbf{x}} + \mathbf{C}\dot{\mathbf{x}} + \mathbf{K}\mathbf{x} + \mathbf{C}_n g_c(\mathbf{x}, \dot{\mathbf{x}}) + \mathbf{K}_n g_k(\mathbf{x}, \dot{\mathbf{x}}) = \mathbf{f}
```
where $\mathbf{\Xi}_n g_{\bullet}(\mathbf{x},\dot{\mathbf{x}}) = \mathbf{C}_n g_c(\mathbf{x}, \dot{\mathbf{x}}) + \mathbf{K}_n g_k(\mathbf{x}, \dot{\mathbf{x}})$ represents the nonlinear system forces. For example, a 3DOF Duffing oscillator, connected at one end, would have representative nonlinear forces,
```math
\mathbf{K}_n g_n(\mathbf{x}) = \begin{bmatrix}
    k_{n,1} & - k_{n,2} & 0 \\
    0 & k_{n,2} & -k_{n,3} \\
    0 & 0 & k_{n,3}
\end{bmatrix}
\begin{bmatrix}
    x_1^3 \\
    (x_2-x_1)^3 \\
    (x_3 - x_2)^3
\end{bmatrix}
```

## Installing DynaSim

To install DynaSim, follow these steps:

Linux and macOS:
```
python3 -m pip install dynasim
```

Windows:
```
py -m pip install dynasim
```
## Using DynaSim

### Quickstart Guide

To use DynaSim, here is a quick start guide to generate a 5-DOF oscillating system with some cubic stiffness nonlinearities:

```
import numpy as np
import dynasim

# create required variables
n_dof = 5

# create a time vector of 2048 time points up to 120 seconds
nt = 2048
time_span = np.linspace(0, 120, nt)

# create vectors of system parameters for sequential MDOF
m_vec = 10.0 * np.ones(n_dof)
c_vec = 1.0 * np.ones(n_dof)
k_vec = 15.0 * np.ones(n_dof)
# imposes every other connection as having an additional cubic stiffness
kn_vec = np.array([25.0 * (i%2) for i in range(n_dof)])

# create nonlinearities
system_nonlin = dynasim.nonlinearities.exponent_stiffness(kn_vec, exponent=3, dofs=n_dof)
# instantiate system and embed nonlinearity
system = dynasim.systems.mdof_cantilever(m_vec, c_vec, k_vec, dofs=n_dof, nonlinearity=system_nonlin)

# create excitations and embed to system
system.excitations = [None] * n_dof
system.excitations[-1] = dynasim.actuators.sine_sweep(w_l = 0.5, w_u = 2.0, F0 = 1.0)

# simulate system
data = system.simulate(time_span, z0=None)
```

### Nonlinearities

Three nonlinearities are available, exponent stiffness, exponent damping, and Van der Pol damping

```
dynasim.nonlinearities.exponent_stiffness(kn_vec, exponent=3, dofs=n_dof)
dynasim.nonlinearities.exponent_damping(cn_vec, exponent=0.5, dofs=n_dof)
dynasim.nonlinearities.vanDerPol(cn_vec, dofs=n_dof)
```
These classes contain the $g_k(\mathbf{x}, \dot{\mathbf{x}})$ function.

### Common system classes

There are a currently two system types available for MDOF systems, which are instantiated from vectors of system parameter values:

```
dynasim.systems.mdof_symmetric(m_, c_, k_, dofs, nonlinearity)
dynasim.systems.mdof_cantilever(m_, c_, k_, dofs, nonlinearity)
```

There are also two grid array mass systems, which simulate the MDOF systems with x and y directions. The displacement vectors are now $\{x_{11},y_{11},x_{12},y_{12}...,x_{mn},y_{mn}\} $, and the corresponding velocity and state vectors. These are available as coupled or uncoupled, i.e. whether the restoring force take into account relative angles between nodes as well. When using the coupled system, make sure to use the `corotational_rk4` simulator.

```
dynasim.systems.grid_uncoupled(m_vec, ch_mat, cv_mat, kh_mat, kv_mat)
dynasim.systems.grid_corotational(m_vec, ch_mat, cv_mat, kh_mat, kv_mat)
```

### Actuator classes

The forcing for the system should be a list of actuations, equal in length to the number of DOFs of the system, there many actuation types,
```
dynasim.actuators.sinusoid(...)
dynasim.actuators.white_gaussian(...)
dynasim.actuators.sine_sweep(...)
dynasim.actuators.rand_phase_ms(...)
dynasim.actuators.banded_noise(...)
```

### Totally custom system

One can generate a custom system by instantiating an MDOF system with corresponding modal matrices, but the nonlinearity must also be instantiated and 
```
dynasim.base.mdof_system(M, C, K, Cn, Kn)
```

## Continuous Beams

These work much like the MDOF systems but with a few tweaks. Where ```beam_kwargs_cmb``` is a dictionary of combined properties of the beam.
```
beam_kwargs_cmb = {
    "EI" : EI,  # Young's Modulus multiplied by second moment of intertia
    "pA" : pA,  # density multiplied by cross sectional area
    "c" : c,  # damping
    "l" : l  # length of beam
}
ss_beam = dynasim.systems.cont_beam("cmb_vars", **beam_kwargs_cmb)
```

One can then generate and retrieve the mode shapes of the beam via
```
xx, phis = ss_beam.gen_modes(support, n_modes, nx)
```
Where ```support``` is a string representing different support types for the beam, options available are:
| Code | Description |
|------|-------------|
| `ss-ss` | Simply supported at both ends |
| `fx-fx` | Fixed at both ends |
| `fr-fr` | Free at both ends |
| `fx-ss` | Fixed at one end and simply supported at the other |
| `fx-fr` | Fixed at one end and free at the other |

Then, excitations are added as a list of dictionaries containing the location (in m) and excitation (prescribed as with mdof systems)
```
ss_beam.excitations = [
    {
        "excitation" : dynasim.actuators.sinusoid(1.2, 1.0),
        "loc" : 0.25
    }
]
```
Then simulate just as with the MDOF system, however, the returned data here is the modal coordinates, and so to retrieve displacement, simply multiply by, and sum through, the mode shapes.
```
data = ss_beam.simulate(tt, z0 = tau0)

ww = np.sum(np.matmul(data['x'][:,:,np.newaxis], phis.T[:,np.newaxis,:]), axis=0)
wwd = np.sum(np.matmul(data['xdot'][:,:,np.newaxis], phis.T[:,np.newaxis,:]), axis=0)
```

## Contact

If you want to contact me you can reach me at <mhaywood@ethz.ch>.

## License
<!--- If you're not sure which open license to use see https://choosealicense.com/--->

This project uses the following license: [MIT](<https://opensource.org/license/mit/>).

#' Compute state transition probability matrix for a given lag
#'
#' @param x Numeric vector, the time series.
#' @param n_s Integer, number of states (quantile bins). Default 15.
#' @param lag Integer, lag order for transition. Default 1.
#'
#' @return A matrix of dimension n_s x n_s with row-wise transition probabilities.
#' @export
#'
#' @examples
#' x <- arima.sim(model = list(ar = 0.5), n = 1000)
#' P <- transition_matrix(x, n_s = 15, lag = 1)
transition_matrix <- function(x, n_s = 15, lag = 1) {
  states <- discretize_quantile(x, n_s)
  n <- length(states)
  if (n <= lag) stop("Time series too short for the specified lag")

  from <- states[1:(n - lag)]
  to   <- states[(1 + lag):n]

  counts <- table(from, to)
  P <- matrix(0, n_s, n_s)
  row_idx <- as.integer(rownames(counts))
  col_idx <- as.integer(colnames(counts))
  for (i in seq_along(row_idx)) {
    P[row_idx[i], col_idx] <- counts[i, ]
  }

  row_sums <- rowSums(P)
  P[row_sums > 0, ] <- P[row_sums > 0, ] / row_sums[row_sums > 0]
  return(P)
}

#' Multivariate DSC (MDSC) Input: X and Y are matrices where columns are components
#'
#' @param x Numeric vector, first component.
#' @param y Numeric vector, second component.
#' @param n_s Integer, number of states.
#' @param lag Integer, lag order.
#'
#' @return Matrix of dimension n_s x n_s.
#' @export
#'
cross_transition_matrix <- function(x, y, n_s = 15, lag = 1) {
  states_x <- discretize_quantile(x, n_s)
  states_y <- discretize_quantile(y, n_s)
  n <- length(states_x)
  if (n <= lag) stop("Time series too short for specified lag")

  from <- states_x[1:(n - lag)]
  to   <- states_y[(1 + lag):n]

  counts <- table(from, to)
  P <- matrix(0, n_s, n_s)
  row_idx <- as.integer(rownames(counts))
  col_idx <- as.integer(colnames(counts))
  for (i in seq_along(row_idx)) {
    P[row_idx[i], col_idx] <- counts[i, ]
  }
  row_sums <- rowSums(P)
  P[row_sums > 0, ] <- P[row_sums > 0, ] / row_sums[row_sums > 0]
  return(P)
}

#' Quantile discretization and transition probability matrix
#'
#' @param x Numeric vector representing the time series.
#' @param n_s Integer, number of states (quantile bins). Default is 15.
#'
#' @return Integer vector of state labels (1 to n_s).
#' @export
#'
#' @examples
#' x <- rnorm(1000)
#' states <- discretize_quantile(x, n_s = 10)
discretize_quantile <- function(x, n_s = 15) {
  if (n_s < 2) stop("n_s must be at least 2")
  probs <- seq(0, 1, length.out = n_s + 1)[-c(1, n_s + 1)]
  qs <- stats::quantile(x, probs = probs, na.rm = TRUE)
  breaks <- c(-Inf, qs, Inf)
  as.integer(cut(x, breaks = breaks, labels = FALSE))
}

#' Compute pairwise GDSC matrix for a collection of series
#'
#' @param X Matrix with time series in columns (rows = time points).
#' @param n_s Integer, number of states.
#' @param lags Vector of lag orders.
#' @param th_step Numeric, threshold step size.
#' @param show_progress Logical.
#'
#' @return Matrix of GDSC values between columns.
#' @export
#'
#' @examples
#' n <- 1000
#' X <- matrix(0,n,4)
#' X[,1]=arima.sim(model = list(ar = 0.5), n = n)
#' X[,2]=arima.sim(model = list(ar = 0.5), n = n)
#' X[,3]=arima.sim(model = list(ar = -0.5), n = n)
#' X[,4]=arima.sim(model = list(ar = -0.5), n = n)
#' gdsc_matrix(X, n_s = 10,lags=1:3,th_step = 0.005)
gdsc_matrix <- function(X, n_s = 15, lags = 1:5, th_step = 0.005,
                        show_progress = FALSE) {
  X <- as.matrix(X)
  m <- ncol(X)
  mat <- matrix(0, m, m)
  colnames(mat) <- rownames(mat) <- colnames(X)

  if (show_progress) pb <- utils::txtProgressBar(min = 0, max = m*(m-1)/2, style = 3)
  counter <- 0
  for (i in 1:m) {
    for (j in i:m) {
      val <- gdsc(X[, i], X[, j], n_s, lags, th_step)
      mat[i, j] <- val
      mat[j, i] <- val
      if (i != j) {
        counter <- counter + 1
        if (show_progress) utils::setTxtProgressBar(pb, counter)
      }
    }
  }
  if (show_progress) close(pb)
  return(mat)
}

#' Global DSC (GDSC) across multiple lags
#'
#' @param x Numeric vector, first time series.
#' @param y Numeric vector, second time series.
#' @param n_s Integer, number of states.
#' @param lags Vector of lag orders to include.
#' @param th_step Numeric, threshold step size.
#'
#' @return Numeric, average DSC over specified lags.
#' @export
#' @references
#' Nie, Chun-Xiao. "Distinguishing time series generated by different intrinsic dynamics using quantile graphs." Communications in Nonlinear Science and Numerical Simulation (2026): 110002.
#' @examples
#' x <- arima.sim(model = list(ar = c(0.5, -0.3)), n = 1000)
#' y <- arima.sim(model = list(ar = c(0.5, -0.3)), n = 1000)
#' gdsc(x, y, n_s = 15, lags = 1:3)
gdsc <- function(x, y, n_s = 15, lags = 1:5, th_step = 0.005) {
  vals <- vapply(lags, function(l) dsc(x, y, n_s, l, th_step), numeric(1))
  mean(vals)
}

#' Compute Dynamics Similarity Coefficient (DSC) between two time series
#'
#' @param x Numeric vector, first time series.
#' @param y Numeric vector, second time series.
#' @param n_s Integer, number of states. Default 15.
#' @param lag Integer, lag order. Default 1.
#' @param th_step Numeric, threshold step size. Default 0.005.
#'
#' @return A numeric value between 0 and 1, where higher values indicate greater similarity of intrinsic dynamics.
#' @export
#'
#' @references
#' Nie, Chun-Xiao. "Distinguishing time series generated by different intrinsic dynamics using quantile graphs." Communications in Nonlinear Science and Numerical Simulation (2026): 110002.
#' @examples
#' x <- arima.sim(model = list(ar = 0.5), n = 1000)
#' y <- arima.sim(model = list(ar = -0.5), n = 1000)
#' dsc(x, y, n_s = 15, lag = 1)
dsc <- function(x, y, n_s = 15, lag = 1, th_step = 0.005) {
  Px <- transition_matrix(x, n_s, lag)
  Py <- transition_matrix(y, n_s, lag)

  max_th <- 1 / n_s
  thresholds <- seq(0, max_th, by = th_step)

  jaccards <- vapply(thresholds, function(th) {
    Ax <- (Px > th) * 1      # strict inequality as requested
    Ay <- (Py > th) * 1
    inter <- sum(Ax & Ay)
    union <- sum(Ax | Ay)
    if (union == 0) 1 else inter / union
  }, numeric(1))

  mean(jaccards)
}

#' Compute Multivariate Dynamics Similarity Coefficient (MDSC)
#'
#' @param X Matrix or data.frame, each column is a component time series.
#' @param Y Matrix or data.frame, each column is a component time series.
#' @param n_s Integer, number of states.
#' @param lag Integer, lag order.
#' @param th_step Numeric, threshold step size.
#'
#' @return Numeric, MDSC value.
#' @export
#' @references
#' Nie, Chun-Xiao. "Distinguishing time series generated by different intrinsic dynamics using quantile graphs." Communications in Nonlinear Science and Numerical Simulation (2026): 110002.
#' @examples
#' X <- matrix(rnorm(3000), ncol = 3)
#' Y <- matrix(rnorm(3000), ncol = 3)
#' mdsc(X, Y, n_s = 10)
mdsc <- function(X, Y, n_s = 15, lag = 1, th_step = 0.005) {
  X <- as.matrix(X)
  Y <- as.matrix(Y)
  if (ncol(X) != ncol(Y)) stop("X and Y must have the same number of columns")
  k <- ncol(X)

  dsc_vals <- matrix(0, k, k)
  for (i in 1:k) {
    for (j in 1:k) {
      if (i == j) {
        Px <- transition_matrix(X[, i], n_s, lag)
        Py <- transition_matrix(Y[, j], n_s, lag)
      } else {
        Px <- cross_transition_matrix(X[, i], X[, j], n_s, lag)
        Py <- cross_transition_matrix(Y[, i], Y[, j], n_s, lag)
      }
      max_th <- 1 / n_s
      thresholds <- seq(0, max_th, by = th_step)
      jaccards <- vapply(thresholds, function(th) {
        Ax <- (Px > th) * 1      # strict inequality as requested
        Ay <- (Py > th) * 1
        inter <- sum(Ax & Ay)
        union <- sum(Ax | Ay)
        if (union == 0) 1 else inter / union
      }, numeric(1))
      dsc_vals[i, j] <- mean(jaccards)
    }
  }
  mean(dsc_vals)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/discretize_quantile.R
\name{discretize_quantile}
\alias{discretize_quantile}
\title{Quantile discretization and transition probability matrix}
\usage{
discretize_quantile(x, n_s = 15)
}
\arguments{
\item{x}{Numeric vector representing the time series.}

\item{n_s}{Integer, number of states (quantile bins). Default is 15.}
}
\value{
Integer vector of state labels (1 to n_s).
}
\description{
Quantile discretization and transition probability matrix
}
\examples{
x <- rnorm(1000)
states <- discretize_quantile(x, n_s = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dsc.R
\name{dsc}
\alias{dsc}
\title{Compute Dynamics Similarity Coefficient (DSC) between two time series}
\usage{
dsc(x, y, n_s = 15, lag = 1, th_step = 0.005)
}
\arguments{
\item{x}{Numeric vector, first time series.}

\item{y}{Numeric vector, second time series.}

\item{n_s}{Integer, number of states. Default 15.}

\item{lag}{Integer, lag order. Default 1.}

\item{th_step}{Numeric, threshold step size. Default 0.005.}
}
\value{
A numeric value between 0 and 1, where higher values indicate greater similarity of intrinsic dynamics.
}
\description{
Compute Dynamics Similarity Coefficient (DSC) between two time series
}
\examples{
x <- arima.sim(model = list(ar = 0.5), n = 1000)
y <- arima.sim(model = list(ar = -0.5), n = 1000)
dsc(x, y, n_s = 15, lag = 1)
}
\references{
Nie, Chun-Xiao. "Distinguishing time series generated by different intrinsic dynamics using quantile graphs." Communications in Nonlinear Science and Numerical Simulation (2026): 110002.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdsc_matrix.R
\name{gdsc_matrix}
\alias{gdsc_matrix}
\title{Compute pairwise GDSC matrix for a collection of series}
\usage{
gdsc_matrix(X, n_s = 15, lags = 1:5, th_step = 0.005, show_progress = FALSE)
}
\arguments{
\item{X}{Matrix with time series in columns (rows = time points).}

\item{n_s}{Integer, number of states.}

\item{lags}{Vector of lag orders.}

\item{th_step}{Numeric, threshold step size.}

\item{show_progress}{Logical.}
}
\value{
Matrix of GDSC values between columns.
}
\description{
Compute pairwise GDSC matrix for a collection of series
}
\examples{
n <- 1000
X <- matrix(0,n,4)
X[,1]=arima.sim(model = list(ar = 0.5), n = n)
X[,2]=arima.sim(model = list(ar = 0.5), n = n)
X[,3]=arima.sim(model = list(ar = -0.5), n = n)
X[,4]=arima.sim(model = list(ar = -0.5), n = n)
gdsc_matrix(X, n_s = 10,lags=1:3,th_step = 0.005)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdsc.R
\name{gdsc}
\alias{gdsc}
\title{Global DSC (GDSC) across multiple lags}
\usage{
gdsc(x, y, n_s = 15, lags = 1:5, th_step = 0.005)
}
\arguments{
\item{x}{Numeric vector, first time series.}

\item{y}{Numeric vector, second time series.}

\item{n_s}{Integer, number of states.}

\item{lags}{Vector of lag orders to include.}

\item{th_step}{Numeric, threshold step size.}
}
\value{
Numeric, average DSC over specified lags.
}
\description{
Global DSC (GDSC) across multiple lags
}
\examples{
x <- arima.sim(model = list(ar = c(0.5, -0.3)), n = 1000)
y <- arima.sim(model = list(ar = c(0.5, -0.3)), n = 1000)
gdsc(x, y, n_s = 15, lags = 1:3)
}
\references{
Nie, Chun-Xiao. "Distinguishing time series generated by different intrinsic dynamics using quantile graphs." Communications in Nonlinear Science and Numerical Simulation (2026): 110002.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transition_matrix.R
\name{transition_matrix}
\alias{transition_matrix}
\title{Compute state transition probability matrix for a given lag}
\usage{
transition_matrix(x, n_s = 15, lag = 1)
}
\arguments{
\item{x}{Numeric vector, the time series.}

\item{n_s}{Integer, number of states (quantile bins). Default 15.}

\item{lag}{Integer, lag order for transition. Default 1.}
}
\value{
A matrix of dimension n_s x n_s with row-wise transition probabilities.
}
\description{
Compute state transition probability matrix for a given lag
}
\examples{
x <- arima.sim(model = list(ar = 0.5), n = 1000)
P <- transition_matrix(x, n_s = 15, lag = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cross_transition_matrix.R
\name{cross_transition_matrix}
\alias{cross_transition_matrix}
\title{Multivariate DSC (MDSC) Input: X and Y are matrices where columns are components}
\usage{
cross_transition_matrix(x, y, n_s = 15, lag = 1)
}
\arguments{
\item{x}{Numeric vector, first component.}

\item{y}{Numeric vector, second component.}

\item{n_s}{Integer, number of states.}

\item{lag}{Integer, lag order.}
}
\value{
Matrix of dimension n_s x n_s.
}
\description{
Multivariate DSC (MDSC) Input: X and Y are matrices where columns are components
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdsc.R
\name{mdsc}
\alias{mdsc}
\title{Compute Multivariate Dynamics Similarity Coefficient (MDSC)}
\usage{
mdsc(X, Y, n_s = 15, lag = 1, th_step = 0.005)
}
\arguments{
\item{X}{Matrix or data.frame, each column is a component time series.}

\item{Y}{Matrix or data.frame, each column is a component time series.}

\item{n_s}{Integer, number of states.}

\item{lag}{Integer, lag order.}

\item{th_step}{Numeric, threshold step size.}
}
\value{
Numeric, MDSC value.
}
\description{
Compute Multivariate Dynamics Similarity Coefficient (MDSC)
}
\examples{
X <- matrix(rnorm(3000), ncol = 3)
Y <- matrix(rnorm(3000), ncol = 3)
mdsc(X, Y, n_s = 10)
}
\references{
Nie, Chun-Xiao. "Distinguishing time series generated by different intrinsic dynamics using quantile graphs." Communications in Nonlinear Science and Numerical Simulation (2026): 110002.
}
